/*
 * asm1825 - Assembler for PIC16F1825 and others.
 *
 * $Id: eval.c,v 1.2 2017-08-12 00:30:26-04 andrew Exp andrew $
 * 
 * Copyright (c) 2017 Andrew Turnquist
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 * 
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "eval.h"
#include "lc.h"
#include "opcodes.h"
#include "symtab.h"
#include "util.h"
#include "psect.h"


#define FREE(x) if (x) { free(x); x = NULL; }

extern int pass;


void evalUnary(char **tokv)
{
  int i;
  int x = 0;
  int a;
  char buf[64];

  i = 0;
  while (tokv[i]) {

    // Make sure we've evaluating an operator and not a number
    if (tokIsNumber(tokv[i])) { i++; continue; }

    if ((tokv[i][0] == '+') || (tokv[i][0] == '-')) {
      MESSAGE(MSG_DEBUG, "%d: evalUnary", i);

      // Skip if left token is a label or number
      if (i && tokIsLabel(tokv[i-1])) { i++; continue; }
      if (i && tokIsNumber(tokv[i-1])) { i++; continue; }

      // Must have argument on right
      if (tokv[i+1] == NULL) {
        fprintf(stderr, "Expression syntax error near '%s'\n", tokv[i]);
        exit(1);
      }

      // Parse out the number
      // QUESTION: Is this still needed? Does evalString alreay
      //           parse numbers?
      a = parseNumber(tokv[i+1]);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]);
        exit(1);
      }
      if (tokv[i][0] == '+') x = a;
      if (tokv[i][0] == '-') x = -a;
      FREE(tokv[i]);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i] = strdup(buf);

      tokShiftL(tokv, i+1, 1);
      i--;
    }
    i++;
  }
  MESSAGE(MSG_DEBUG, "Leaving evalUnary()");
}


void evalMulDiv(char **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  i = 0;
  while (tokv[i]) {
    if ((tokv[i][0] == '*') || (tokv[i][0] == '/') || (tokv[i][0] == '%')) {
      // fprintf(stderr, "%d: Mult or div\n", i);
      // Evaluate multiplication or division
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]);
      b = parseNumber(tokv[i+1]);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]);
        exit(1);
      }
      if (tokv[i][0] == '*') x = a * b;
      if (tokv[i][0] == '/') x = a / b;
      if (tokv[i][0] == '%') x = a % b;
      FREE(tokv[i-1]);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1] = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}


void evalAddSub(char **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  // fprintf(stderr, "***** evalAddSub: tokv before eval:\n");
  // dumptokens(tokv);
  i = 0;
  while (tokv[i]) {
    // Make sure we've evaluating an operator and not a number
    if (tokIsNumber(tokv[i])) { i++; continue; }

    // fprintf(stderr, "*** evalAddSub: evaluating token %d: '%s'\n", i, tokv[i]);
    if ((tokv[i][0] == '+') || (tokv[i][0] == '-')) {
      // fprintf(stderr, "%d: Add or sub\n", i);
      // Evaluate addition or subtraction
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]);
      b = parseNumber(tokv[i+1]);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]);
        exit(1);
      }
      if (tokv[i][0] == '+') x = a + b;
      if (tokv[i][0] == '-') x = a - b;
      // fprintf(stderr, "Freeing left token (%s)\n", tokv[i-1]);
      assert(tokv[i-1]);
      FREE(tokv[i-1]);
      snprintf(buf,sizeof(buf), "%d", x);
      // fprintf(stderr, "Setting left token to result\n");
      tokv[i-1] = strdup(buf);

      // fprintf(stderr, "Before tokShiftL:\n"); tokDump(tokv);
      tokShiftL(tokv, i, 2);
      // fprintf(stderr, "After tokShiftL:\n"); tokDump(tokv);
      i--;
    }
    i++;
  }
}


void evalBitShift(char **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  // fprintf(stderr, "***** evalBitShift: tokv before eval:\n");
  // dumptokens(tokv);
  i = 0;
  while (tokv[i]) {
    if ((!strcmp(tokv[i], "<<")) || (!strcmp(tokv[i], ">>"))) {
      // Evaluate addition or subtraction
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]);
      b = parseNumber(tokv[i+1]);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]);
        exit(1);
      }
      if (!strcmp(tokv[i], "<<")) x = a << b;
      if (!strcmp(tokv[i], ">>")) x = a >> b;
      // fprintf(stderr, "Freeing left token (%s)\n", tokv[i-1]);
      assert(tokv[i-1]);
      FREE(tokv[i-1]);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1] = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}


void evalBitAnd(char **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  // fprintf(stderr, "***** evalBitAnd: tokv before eval:\n");
  // dumptokens(tokv);
  i = 0;
  while (tokv[i]) {
    if (tokv[i][0] == '&') {
      // Evaluate bitwise AND
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]);
      b = parseNumber(tokv[i+1]);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]);
        exit(1);
      }
      if (tokv[i][0] == '&') x = a & b;
      assert(tokv[i-1]);
      FREE(tokv[i-1]);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1] = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}


void evalBitXor(char **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  // fprintf(stderr, "***** evalBitXor: tokv before eval:\n");
  // dumptokens(tokv);
  i = 0;
  while (tokv[i]) {
    if (tokv[i][0] == '^') {
      // Evaluate bitwise XOR
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]);
      b = parseNumber(tokv[i+1]);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]);
        exit(1);
      }
      if (tokv[i][0] == '^') x = a ^ b;
      assert(tokv[i-1]);
      FREE(tokv[i-1]);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1] = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}


void evalBitOr(char **tokv)
{
  int i;
  int x = 0;
  int a,b;
  char buf[64];

  // fprintf(stderr, "***** evalBitOr: tokv before eval:\n");
  // dumptokens(tokv);
  i = 0;
  while (tokv[i]) {
    if (tokv[i][0] == '|') {
      // Evaluate bitwise OR
      if ((i == 0) || (tokv[i+1] == NULL)) {
        fprintf(stderr, "Expression syntax error\n");
        exit(1);
      }
      a = parseNumber(tokv[i-1]);
      b = parseNumber(tokv[i+1]);
      if (a == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i-1]);
        exit(1);
      }
      if (b == EVAL_FAIL) {
        message(MSG_ERROR, "Failed to evaluate number '%s'", tokv[i+1]);
        exit(1);
      }
      if (tokv[i][0] == '|') x = a | b;
      assert(tokv[i-1]);
      FREE(tokv[i-1]);
      snprintf(buf,sizeof(buf), "%d", x);
      tokv[i-1] = strdup(buf);

      tokShiftL(tokv, i, 2);
      i--;
    }
    i++;
  }
}



void evalParens(char **tokv)
{
  int i;
  int pstart = -1;
  int pdepth = 0;
  char **subv;

  i=0;
  while (tokv[i]) {
    if (tokv[i][0] == '(') {
      // Begin parenthesis evaluation
      // fprintf(stderr, "----> Open Paren\n");
      if (pstart < 0) pstart = i;
      pdepth++;
    }
    if (tokv[i][0] == ')') {
      // End parenthesis evaluation
      // fprintf(stderr, "----> End Paren\n");
      if (pstart < 0) {
        fprintf(stderr, "ERROR: Syntax error: close paren with matching open paren.\n");
        exit(1);
      }
      pdepth--;
      if (pdepth == 0) {
        /* Evaluate from pstart to here */
        subv = tokDup(tokv, pstart+1, i-1);
        // fprintf(stderr, "(2) Sub-clause to evaluate:\n");
        // dumptokens(subv);
        evalParens(subv);
	evalUnary(subv);
        evalMulDiv(subv);
        evalAddSub(subv);
        evalBitShift(subv);
        evalBitAnd(subv);
        evalBitXor(subv);
        evalBitOr(subv);
        /* Expression should have evaluated down to one value, we hope. */
        tokv[pstart] = strdup(subv[0]);
        tokShiftL(tokv, pstart+1, (i-(pstart)));
        FREE(subv[0]);
        pstart = -1;    // Reset to no open paren
      }  // pdepth == 0
    }  // end paren
    i++;
  }
}


int evalString(const char *str)
{
  char **tokv;
  int i;
  int x;
  char xstr[32];

  MESSAGE(MSG_DEBUG, "Evaluating string '%s'", str);
  tokv = tokSplit(str);

  i=0;
  while (tokv[i]) {
    if (tokIsLabel(tokv[i])) {
      if (!isdigit(tokv[i][0])) {
        // fprintf(stderr, "Resolving label or literal '%s'\n", tokv[i]);
        if (!strcmp(tokv[i], "$")) {
          // special 'here' label
          MESSAGE(MSG_INFO, "evalString(): Evaluating $ to 0x%04x",
              (pass ? psAddrBase(curpsect->pnum) : 0) + lcGet());
          snprintf(xstr, sizeof(xstr), "%d",
              (pass ? psAddrBase(curpsect->pnum) : 0) + lcGet());
          FREE(tokv[i]);
          tokv[i] = strdup(xstr);
        } else {
          // Symbol to lookup
          x = symLookup(tokv[i]);
          if (x != EVAL_FAIL) {
            snprintf(xstr, sizeof(xstr), "%d", x);
            FREE(tokv[i]);
            tokv[i] = strdup(xstr);
          } else {
            // If a symbol couldn't resolve, return error so it gets
            // marked as a forward reference to (hopefully) be resolved
            // during pass 2.
            //
            tokFree(tokv);
            return EVAL_FAIL;
          }  // if/else symLookup resolved symbol
        }  // if/else 'here' label
      }  // not number
    }
    i++;
  }  // while (label resolving)

  // fprintf(stderr, "Initial tokv[] array:\n");
  // tokDump(tokv);
  // printf("    Total tokens: %d\n", tokCount(tokv));
  evalParens(tokv);
  evalUnary(tokv);
  evalMulDiv(tokv);
  evalAddSub(tokv);
  evalBitShift(tokv);
  evalBitAnd(tokv);
  evalBitXor(tokv);
  evalBitOr(tokv);
  // printf("After evals: %d\n", tokCount(tokv));
  // sscanf(tokv[0], "%i", &x);
  return parseNumber(tokv[0]);
}


